/************************************************************************

 The contents of this file are subject to the Mozilla Public License
 Version 1.1 (the "License"); you may not use this file except in	
 compliance with the License. You may obtain a copy of the License at
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS"
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 License for the specific language governing rights and limitations
 under the License.

 The Original Code is Isearch.

 The Initial Developer of the Original Code is Clearinghouse for
 Networked Information Discovery and Retrieval.

 Portions created by Netscape Communications are Copyright (C) 2001. 
 All Rights Reserved.

 Portions created by the Clearinghouse for Networked Information
 Discovery and Retrieval are Copyright (C) 1994. All Rights Reserved.

************************************************************************/

/* Parses fields based on colon delimeted lines. <TAG>:<TEXT> */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include "site.h"


SITEDOC::SITEDOC (PIDBOBJ DbParent): DOCTYPE (DbParent) {}


GPTYPE 
SITEDOC::ParseWords(unsigned char* DataBuffer, // Pointer to document text buffer.
		    int DataLength,            // Length of document text buffer in # of characters.
		    int DataOffset,            // Offset that must be added to all GP positions 
		                               // because GP space is shared with other documents. 
		    unsigned int FileId,       // file identification number
		    int *fdList)               // list of open files to write data to
{     

    unsigned int field = 0;
    unsigned int f = 0;
    unsigned int fid = 0;
    index_entry ie;
    int fn;
    int numAdded = 0;
    int scanning_for_tag = 0;
    int Position = 0;
    int linenum = 0;
    unsigned char utf8_char[8];
    int i;
    int break_after_this_utf_char;
    
    // unsigned char tag;
    // tag = DataBuffer[Position];

    // get the colon delimited field (an ascii character)
    if (Position < DataLength)
        field = DataBuffer[Position];

    // get past the colon tag 
    while ((Position < DataLength) && (DataBuffer[Position] != ':'))
	Position++;
    
    if (Position < DataLength)
	Position++;
    
    while (Position < DataLength) {
	// scan to the begining of the word
	
	break_after_this_utf_char = 0;
	
	while ((Position < DataLength) &&
	       (!IsAlnum(DataBuffer[Position])) &&
	       (!IsUTF8_Start_Char(DataBuffer[Position]))) {
	    
	    if (DataBuffer[Position] == '\n') {
		linenum++;

		// get the colon delimited field (an ascii character)
		while ((Position < DataLength) && (DataBuffer[Position] == '\n'))
		  Position++;

		if (Position < DataLength)
		  field = DataBuffer[Position];

		// get past the tag 
		while ((Position < DataLength) && (DataBuffer[Position] != ':'))
		    Position++;
	    }
	    
	    Position++;
	}
	
	// remember this offset as the beginning of a word
	if (Position < DataLength) {

	    fn = GetBucket(DataBuffer[Position]);
	  
	    f = FIELD_MASK & (field << 24);
	    if ( FileId > FILEID_MASK ) {
	        fprintf(stderr,"ERROR: FileId %d has exeeded the maximum allowable value %d.\n",FileId, FILEID_MASK);
		fid = FileId;
	    }
	    else {
	        fid = FileId | f;
	    }
	  
	    ie.gpindex = DataOffset + Position;
	    ie.mdtindex = fid;
	    
	    if ( sizeof(index_entry) != write(fdList[fn], (void *)&ie, sizeof(index_entry))) {
	        fprintf(stderr,"ERROR: can't write to fd %d\n");
		exit(88);
	    }
	    numAdded++;
	
	    // Skip over UTF-8 continuation characters
	    if (IsUTF8_Start_Char(DataBuffer[Position])) {
	      
		i = 0;
		utf8_char[i++] = DataBuffer[Position];
		
		Position++;
		
		while ((Position < DataLength) &&  (i < 6) && 
		       IsUTF8_Continuation_Char(DataBuffer[Position])) {
		    // fprintf(stderr,"%c", DataBuffer[Position]);
		    
		    utf8_char[i++] = DataBuffer[Position];
		    
		    Position++;
		}
		
		utf8_char[i] = 0;
		
		// if this is a japanese/chinese/korean character then break after 
		// this UTF-8 char
		if (IsUTF8_CJK_Char(utf8_char)) {
		    break_after_this_utf_char = 1;
		}
	    }
	    else 
		Position++;
	}
	
	// scan to the end of the word, delimited by a non-alphacharacter 
	while ((Position < DataLength) && 
	       IsAlnum(DataBuffer[Position]) && 
	       !break_after_this_utf_char) {
	    Position++;
	}
	
    }
    return numAdded;
} 

// we prefer records this length or shorter
#define OPTIMAL_RECORD_LENGTH         (80)    
#define MAX_REASONABLE_RECORD_LENGTH  (400)
#define MAX_SPAM_RECORD_LENGTH        (800)

double 
SITEDOC::ScoreHit(IRESULT result, double termweight) {

    double Weight = 1.0;
    double RecordLengthWeight; 
    int RecordLength;

    //
    // We are scoring Sites
    //
    if (result.MiscFields & COOL_FLAG_MASK) {
        // this site is cool
        Weight = 3.0;
    }
	
    //
    // Weight shorter records more than longer records
    //
    RecordLength = result.GlobalFileEnd - result.GlobalFileStart;
	  
    if (RecordLength >= MAX_SPAM_RECORD_LENGTH) 
        RecordLengthWeight = 0.0001;
    else if (RecordLength > 0) {
        // scale the weight to be  between 0 and 1
        if (RecordLength > MAX_REASONABLE_RECORD_LENGTH) {
	    // make the score fall off agressively after exceeding a reasonable record length
	    // score range [0.75, 0.00] 
	    RecordLengthWeight = 0.75 * ((double)((MAX_SPAM_RECORD_LENGTH - RecordLength)) / 
					 (double) (MAX_SPAM_RECORD_LENGTH - MAX_REASONABLE_RECORD_LENGTH));
	}
	else {
	    // a gradual score change for records in the reasonable scoring range
	    // score range [1.0, 0.75] 
	    RecordLengthWeight =  (((double)(MAX_REASONABLE_RECORD_LENGTH * 4 - RecordLength)) / 
				   (double) (MAX_REASONABLE_RECORD_LENGTH * 4));
	}
    }

    if (result.HitCount < 1)
        result.HitCount = 1;

    // adjust score for the number of hits in this record, range [1.0, 2.0)
    Weight *= ((double)2.0 - ((double)1.0 / ((double)result.HitCount)));

    // adjust score for length of record
    Weight *= RecordLengthWeight;  

    // adjust hit for term weighting
    return(termweight * Weight);
}  

//
// TopLevelCat() - given a category name, determin the numerical
//                 category value. This makes category comparisons
//                 much faster
//

unsigned int 
SITEDOC::TopLevelCat(unsigned char *q) {

  unsigned int Cat = TOP_CAT_None;

  switch (q[0]) {
    
  case 'A' :
    if (q[1] == 'r')
      Cat = TOP_CAT_Arts;
    else if (q[1] == 'd')
      Cat = TOP_CAT_Adult;
    else 
      Cat = TOP_CAT_None;
    break;
  case 'B' :
    if (q[1] == 'o')
      Cat = TOP_CAT_Bookmarks;
    else if (q[1] == 'u')
      Cat = TOP_CAT_Business;
    else 
      Cat = TOP_CAT_None;
    break;
  case 'C' :
    Cat = TOP_CAT_Computers;  
    break;
  case 'G' :
    Cat = TOP_CAT_Games;
    break;
  case 'H' :
    if (q[1] == 'e')
      Cat = TOP_CAT_Health;
    else if (q[1] == 'o')
      Cat = TOP_CAT_Home;
    else 
      Cat = TOP_CAT_None;  
    break;
  case 'K' :
    Cat = TOP_CAT_Kids;
    break;
  case 'N' :
    if (q[2] == 't')
      Cat = TOP_CAT_Netscape;
    else if (q[2] == 'w')
      Cat = TOP_CAT_News;
    else
      Cat = TOP_CAT_None;  
    break;
  case 'R' :
    if (q[2] == 'g')
      Cat = TOP_CAT_Regional;
    else if (q[2] == 'f')
      Cat = TOP_CAT_Reference;
    else if (q[2] == 'c')
      Cat = TOP_CAT_Recreation;
    else 
      Cat = TOP_CAT_None;  
    break;
  case 'S' :
    if (q[1] == 'p')
      Cat = TOP_CAT_Sports;
    else if (q[1] == 'h')
      Cat = TOP_CAT_Shopping;
    else if (q[1] == 'o')
      Cat = TOP_CAT_Society;
    else if (q[1] == 'c')
      Cat = TOP_CAT_Science;
    else 
      Cat = TOP_CAT_None;
    break;
  case 'W' :
    Cat = TOP_CAT_World;
    break;
  default :
    Cat = TOP_CAT_None;
    break;
  }

  return Cat;

}

SITEDOC::~SITEDOC () {}

