/************************************************************************

 The contents of this file are subject to the Mozilla Public License
 Version 1.1 (the "License"); you may not use this file except in	
 compliance with the License. You may obtain a copy of the License at
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS"
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 License for the specific language governing rights and limitations
 under the License.

 The Original Code is Isearch.

 The Initial Developer of the Original Code is Clearinghouse for
 Networked Information Discovery and Retrieval.

 Portions created by Netscape Communications are Copyright (C) 2001. 
 All Rights Reserved.

 Portions created by the Clearinghouse for Networked Information
 Discovery and Retrieval are Copyright (C) 1994. All Rights Reserved.

************************************************************************/

/* Do multi-term search. */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/../string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>

#include "defs.h"
#include "string.h"
#include "common.h"
#include "record.h"
#include "mdtrec.h"
#include "mdt.h"
#include "idbobj.h"
#include "iresult.h"
#include "irset.h"
#include "dtreg.h"
#include "index.h"
#include "idb.h"
#include "catids.h"


void INDEX::ResultRange(index_entry *Index_Table,
			int file_length,
			const STRING& Term, 
			int *result_first,
			int *result_last) 
{
  GPTYPE gp;
  int TermLength;
  int ip, maxip, low, high;
  int best_low, best_high;
  int x, z;
  int first, last;
  int match, nomatch;
  int done;
  GDT_BOOLEAN hit;

  *result_first = -1;
  *result_last  = -1;

  // Set up the searching variables
  maxip = (file_length / sizeof(index_entry)) - 1;
  low = 0;
  high = maxip;
  best_low = low;
  best_high = high;
  ip = high / 2;
  z = 0;
  TermLength = strlen((char *)Term.Buffer);

  // Now, do the binary search on the pointers
  done = 0;    
  hit = GDT_FALSE;
  
  while ( (!done) && (high >= low) ) {
      ip = (low + high) / 2;
      gp = Index_Table[ip].gpindex;

#if 0
      /* for debugging binary search of terms */
      fprintf(stderr,"%7d ",ip);
#endif

      z = Match(Term, TermLength, gp);

      if (z == 0) {
	done = 1;
	hit = GDT_TRUE;
      }
      else if (z < 0) {
	high = ip-1;
	best_high = ip;
      }
      else if (z > 0) {
	low = ip + 1;
	best_low = ip;
      }
    } 
    
    if (!hit) {
      // no hits - return an empty range
      return;
    }
    
    // bracket hits

    // find first matching string
    low = best_low;
    high = ip;
    match = ip;
    nomatch = best_low;
    
    // iterate until we're within 5 records of the first element
    while ( (match - nomatch) > 5 ) {
      first = (low + high) / 2;
      gp = Index_Table[first].gpindex;
      z = Match(Term, TermLength, gp);
      
      if (z == 0) {
	match = first;
	high = first;
      } else {
	nomatch = first;
	low = first + 1;
      }
    }
    
    first = match;
    
    z = 0; 
    while((z == 0) && (first > nomatch)) {
      first--;
      gp = Index_Table[first].gpindex;
      z = Match(Term, TermLength, gp);
    }
    if (z != 0) {
      first++;
    }
    
    // find last matching string
    low = ip;
    high = best_high;
    match = ip;
    nomatch = best_high;
    
    while ( (nomatch - match) > 5 ) {
      last = (high + low) / 2;
      gp = Index_Table[last].gpindex;
      z = Match(Term, TermLength, gp);
      if (z == 0) {
	match = last;
	low = last + 1;
      } else {
	nomatch = last;
	high = last;
      }
    } 
    
    last = match;
    
    z = 0;
    while ((z == 0) && (last < nomatch)) {
      last++;
      gp = Index_Table[last].gpindex;
      z = Match(Term, TermLength, gp);
    } 
    
    if (z != 0) {
      last--;
    }

#if 0
    fprintf(stderr, "##################################################\n");
    for(int i = first - 2; i < last + 2 ; i++) {
      gp = Index_Table[i].gpindex;
      z = Match(Term, TermLength, gp);
      fprintf(stderr, "<%4d>  %d \n", i, z);
    }
    fprintf(stderr, "##################################################\n");
#endif

  *result_first = first;
  *result_last  = last;

}

PIRSET INDEX::MultiTermSearch(const STRING& QueryTerm,
			      int first, 
			      int last,
			      unsigned char field,
			      PIRSET WorkingSet)
{
    static long file_length = 0;
    static index_entry *Index_Table = NULL;     // this is the pointer to the mmapped file
    static int fd = -1;
    static STRING TmpIndexFileName;
    
    int y;
    unsigned char f = 0;
    unsigned char OrigTerm[StringCompLength+1], *Term;
    int TermLength, OrigTermLength;
    PIRSET pirset = NULL;
    int ip, w;
    unsigned char Buffer[StringCompLength+1];  
    
    if (first == -1) {
	// no results 
	return NULL;
    }
    
    if ((Index_Table == NULL) || (TmpIndexFileName != IndexFileName)) {
	// Open the index file
	TmpIndexFileName = IndexFileName;
	fd = open(TmpIndexFileName,O_RDONLY);
	if (fd < 0) {
	    perror(TmpIndexFileName);
	    exit(12);
	}
	
	// now mmap the file
	file_length = lseek(fd,0L, SEEK_END);
	Index_Table = (index_entry *) mmap(NULL, file_length ,PROT_READ, MAP_SHARED, fd, 0);
    }
    
    QueryTerm.GetCString((char *)OrigTerm, sizeof(OrigTerm));
    OrigTermLength = strlen((char *)OrigTerm);
    
    if (OrigTermLength < QueryTerm.GetLength()) {
	OrigTerm[OrigTermLength - 1] = '*';
    }
    
    // cerr << "QueryTermLen = " << OrigTermLength << "\n";
    // cerr << "OrigTermlen  = " << strlen(OrigTerm) << "\n";
    
    int PhraseEnd = OrigTermLength;
    if (OrigTerm[OrigTermLength - 1] == '*')
	PhraseEnd--;
  
    Term = OrigTerm; 
    TermLength = OrigTermLength;
    
    // Build result set
    IRESULT iresult;
    MDTREC mdtrec;
    GPTYPE GlobalRecEnd;
    int TermLenNoStar;
  
    if (QueryTerm.GetChr(OrigTermLength) == '*') {
	TermLenNoStar = OrigTermLength - 1; // ignore "*" at end
    } else {
	TermLenNoStar = OrigTermLength;
    }
    
    //
    // Loop through mdt records and add to the irset
    // 
    
    // resize ahead of time
    pirset = new IRSET(Parent, (last - first) + 1);  
    MDT* mainMdt = Parent->GetMainMdt();


    for (ip=first; ip <= last; ip++) {
	
	// only check last character if its not a wildcard search
	if (TermLenNoStar == OrigTermLength) {
	    // make sure the last character is a word seperator
	    GetIndirectBuffer(Index_Table[ip].gpindex, Buffer, 0, StringCompLength);  
	    
	    if (IsAlnum(Buffer[TermLenNoStar])) {
		// the trailing character is not a space, or word seperating character
		continue;
	    }
	}	
	
	/* fetch the mdt information for this record */
	w = Index_Table[ip].mdtindex & FILEID_MASK;
	f = (Index_Table[ip].mdtindex & FIELD_MASK) >> 24;

	/* if the field is set, only include records that match the field value */
	if (field && f && (f != field)) {
	  continue;
	}

	mainMdt->GetEntry(w, &mdtrec);
	
	// cerr << "## GP = " << Index_Table[ip].gpindex
	// << " Index    " << w 
	// << " Top Cat  " << mdtrec.TopLevelCat 
	// << " pathName " << mdtrec.pathName
	// << " Start    " << mdtrec.GlobalFileStart<< "\n";
	
	// make sure that phrases don't go past the end of the local record.
	GlobalRecEnd = mdtrec.GetGlobalFileStart() + mdtrec.GetLocalRecordEnd();
	if ( !((GlobalRecEnd - Index_Table[ip].gpindex) >= (TermLenNoStar - 1)) ) {
	    continue;
	}
	
	// Skip deleted records
	if (mdtrec.MiscFields & DELETED_FLAG_MASK) {
	    continue;
	}
	
	// Check working set to see if we should add this record
	// we only add new instances if the record (site) is already represented,
	// this makes merging faster by reducing the size of the sets, and 
	// avoiding overflow of the IRSETS for comon keywords.
	if (WorkingSet != NULL) {
	    y = WorkingSet->HashGet(w);
	    
	    if (y < 0) {
		// this element isn't in the working set, so skip it
		continue;
	    }
	}
	
	iresult.SetMdtIndex(w);
	iresult.SetHitCount(1);
	iresult.SetScore(0);
	iresult.TopLevelCat     = mdtrec.TopLevelCat;
	iresult.GlobalFileStart = mdtrec.GlobalFileStart;
	iresult.GlobalFileEnd   = mdtrec.GlobalFileEnd;
	iresult.pathName        = mdtrec.pathName;
	iresult.fileName        = mdtrec.fileName;
	iresult.DocumentType    = mdtrec.DocumentType;
	iresult.MiscFields      = mdtrec.MiscFields;
	iresult.CatDepth        = mdtrec.CatDepth;
	iresult.Field           = f;
	
	pirset->AddEntry(iresult, 1);
    }
    
    return pirset;
}


